import logging
from typing import Union

from fs.base import FS
from fs.opener import manage_fs
from schema import Optional, Or

from organize.utils import Template

from ._utils import open_create_fs_path
from .action import Action

logger = logging.getLogger(__name__)

MODES = (
    "prepend",
    "append",
    "overwrite",
)


class Write(Action):

    """Write text to a file.

    If the specified path does not exist it will be created.

    Args:
        text (str):
            The text that should be written. Supports templates.

        file (str):
            The file `text` should be written into. Supports templates.
            Defaults to `organize-out.txt`

        mode (str):
            Can be either `append` (append text to the file), `prepend` (insert text as
            first line) or `overwrite` (overwrite content with text).
            Defaults to `append`.

        newline (str):
            (Optional) Whether to append a newline to the given `text`.
            Defaults to `true`.

        clear_before_first_write (bool):
            (Optional) Clears the file before first appending / prepending text to it.
            This happens only the first time write_file is run. If the rule filters
            don't match anything the file is left as it is.
            Defaults to `false`.

        filesystem (str):
            (Optional) A pyfilesystem opener url of the filesystem the textfile is on.
            If this is not given, the local filesystem is used.
    """

    name = "write"
    arg_schema = Or(
        {
            "text": str,
            Optional("path"): str,
            Optional("mode"): Or(*MODES),
            Optional("newline"): bool,
            Optional("clear_before_first_write"): bool,
            Optional("filesystem"): object,
        },
    )

    def __init__(
        self,
        text: str,
        path: str = "organize-out.txt",
        mode: str = "append",
        newline: bool = True,
        clear_before_first_write: bool = False,
        filesystem: Union[FS, str, None] = None,
    ) -> None:
        self.text = Template.from_string(text)
        self.path = Template.from_string(path)
        self.mode = mode.lower()
        self.clear_before_first_write = clear_before_first_write
        self.newline = newline
        self.filesystem = filesystem or self.Meta.default_filesystem

        self._is_first_write = True

        if self.mode not in MODES:
            raise ValueError("mode must be one of %s" % ", ".join(MODES))

    def pipeline(self, args: dict, simulate: bool):
        text = self.text.render(args)
        path = self.path.render(args)

        dst_fs, dst_path = open_create_fs_path(
            fs=self.filesystem,
            path=path,
            args=args,
            simulate=simulate,
        )

        if self._is_first_write and self.clear_before_first_write:
            self.print(f"Clearing file {dst_path}")
            if not simulate:
                dst_fs.create(dst_path, wipe=True)

        self.print(f'{path}: {self.mode} "{text}"')
        if self.newline:
            text += "\n"

        if not simulate:
            with manage_fs(dst_fs):
                if self.mode == "append":
                    dst_fs.appendtext(dst_path, text)
                elif self.mode == "prepend":
                    content = ""
                    if dst_fs.exists(dst_path):
                        content = dst_fs.readtext(dst_path)
                    dst_fs.writetext(dst_path, text + content)
                elif self.mode == "overwrite":
                    dst_fs.writetext(dst_path, text)

        self._is_first_write = False



from conftest import make_files, read_files

from organize import core

FILES = {
    "test.txt": "",
    "file.txt": "Hello world\nAnother line",
    "another.txt": "",
    "folder": {
        "x.txt": "",
        "empty_sub": {},
    },
    "empty_folder": {},
}


def test_delete_empty_files(testfs):
    config = """
    rules:
      - name: "Recursively delete all empty files"
        locations:
          - path: "/"
        subfolders: true
        filters:
          - empty
        actions:
          - delete
    """
    make_files(testfs, FILES)
    core.run(config, simulate=False, working_dir=testfs)
    result = read_files(testfs)
    assert result == {
        "file.txt": "Hello world\nAnother line",
        "folder": {
            "empty_sub": {},
        },
        "empty_folder": {},
    }


def test_delete_empty_dirs(testfs):
    config = """
    rules:
    - name: "Recursively delete all empty directories"
      locations: "/"
      subfolders: true
      targets: dirs
      filters:
        - empty
      actions:
        - delete
    """
    make_files(testfs, FILES)
    core.run(config, simulate=False, working_dir=testfs)
    result = read_files(testfs)
    assert result == {
        "test.txt": "",
        "file.txt": "Hello world\nAnother line",
        "another.txt": "",
        "folder": {
            "x.txt": "",
        },
    }


def test_delete_deep(testfs):
    files = {
        "files": {
            "folder": {
                "subfolder": {
                    "test.txt": "",
                    "other.pdf": "binary",
                },
                "file.txt": "Hello world\nAnother line",
            },
        }
    }
    make_files(testfs, files)
    config = """
    rules:
      - locations: "."
        actions:
          - delete
      - locations: "."
        targets: dirs
        actions:
          - delete
    """
    # sim
    core.run(config, simulate=True, working_dir=testfs)
    result = read_files(testfs)
    assert result == files

    # run
    core.run(config, simulate=False, working_dir=testfs)
    result = read_files(testfs)
    assert result == {}



from rich.prompt import Prompt
from schema import Optional, Or

from organize import console
from organize.utils import Template

from .action import Action


class Confirm(Action):

    """Ask for confirmation before continuing."""

    name = "confirm"
    schema_support_instance_without_args = True

    arg_schema = Or(
        str,
        {
            Optional("msg"): str,
            Optional("default"): bool,
        },
    )

    def __init__(self, msg="Continue?", default=True):
        self.msg = Template.from_string(msg)
        self.default = default

    def pipeline(self, args: dict, simulate: bool):
        msg = self.msg.render(**args)
        result = console.pipeline_confirm(
            self.get_name(),
            msg,
            default=self.default,
        )
        if not result:
            raise StopIteration("Aborted")

    def __str__(self) -> str:
        return 'Confirm(msg="%s")' % self.msg



import logging
import os
from copy import deepcopy
from datetime import datetime
from typing import Any, List, Sequence, Tuple, Union

import jinja2
from fs import open_fs
from fs import path as fspath
from fs.base import FS
from fs.memoryfs import MemoryFS
from jinja2 import nativetypes


def finalize_placeholder(x):
    # This is used to make the `path` arg available in the filters and actions.
    # If a template uses `path` where no syspath is available this makes it possible
    # to raise an exception.
    if isinstance(x, Exception):
        raise x
    return x


Template = jinja2.Environment(
    variable_start_string="{",
    variable_end_string="}",
    autoescape=False,
    finalize=finalize_placeholder,
)

NativeTemplate = nativetypes.NativeEnvironment(
    variable_start_string="{",
    variable_end_string="}",
    autoescape=False,
    finalize=finalize_placeholder,
)


def basic_args():
    """The basic args which are guaranteed to be available."""
    return {
        "env": os.environ,
        "now": datetime.now(),
        "utcnow": datetime.utcnow(),
    }


class SimulationFS(MemoryFS):
    def __init__(self, fs_url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fs_url = fs_url

    def __str__(self):
        if not self.fs_url:
            return "<Simulation>"
        elif "://" in self.fs_url:
            return "<%s>" % self.fs_url
        return self.fs_url


def open_fs_or_sim(fs_url, *args, simulate=False, **kwargs):
    if simulate:
        simFS = SimulationFS(fs_url)
        return simFS
    return open_fs(fs_url, *args, **kwargs)


def expand_user(fs_url: str) -> str:
    fs_url = os.path.expanduser(fs_url)
    if fs_url.startswith("zip://~"):
        fs_url = fs_url.replace("zip://~", "zip://" + os.path.expanduser("~"))
    elif fs_url.startswith("tar://~"):
        fs_url = fs_url.replace("tar://~", "tar://" + os.path.expanduser("~"))
    return fs_url


def expand_args(template: Union[str, jinja2.environment.Template], args=None):
    if not args:
        args = basic_args()

    if isinstance(template, str):
        text = Template.from_string(template).render(**args)
    else:
        text = template.render(**args)

    # expand user and fill environment vars
    text = expand_user(text)
    text = os.path.expandvars(text)

    return text


def fs_path_from_options(
    path: str, filesystem: Union[FS, str, None] = ""
) -> Tuple[FS, str]:
    """
    path can be a fs_url a normal fs_path
    filesystem is optional and may be a fs_url.

    - user tilde is expanded
    - if a filesystem is given, we use that.
    - otherwise we treat the path as a filesystem.
    """
    path = expand_args(path)

    if filesystem:
        if isinstance(filesystem, str):
            filesystem = expand_args(filesystem)
            return (open_fs(filesystem), path)  # type: ignore
        return (filesystem, path)
    return (open_fs(path), "/")


def unwrap_wrapfs(fs, path):
    from fs.path import abspath
    from fs.wrapfs import WrapFS

    if isinstance(fs, WrapFS):
        fs, path = fs.delegate_path(path)
    return fs, abspath(path)


def is_same_resource(fs1: FS, path1: str, fs2: FS, path2: str):
    from fs.errors import NoSysPath, NoURL
    from fs.tarfs import ReadTarFS, WriteTarFS
    from fs.zipfs import ReadZipFS, WriteZipFS

    # def unwrap_wrapfs(fs, path):
    #     base = "/"
    #     if isinstance(fs, WrapFS):
    #         fs, base = fs.delegate_path("/")
    #     return fs, normpath(join(base, path))  # to support ".." in path

    # completely unwrap WrapFS instances
    fs1, path1 = unwrap_wrapfs(fs1, path1)
    fs2, path2 = unwrap_wrapfs(fs2, path2)

    # obvious check
    if fs1 == fs2 and path1 == path2:
        return True

    # check all fs with syspath support
    try:
        return fs1.getsyspath(path1) == fs2.getsyspath(path2)
    except NoSysPath:
        pass

    # check zip and tar
    Tar = (WriteTarFS, ReadTarFS)
    Zip = (WriteZipFS, ReadZipFS)
    if (isinstance(fs1, Tar) and isinstance(fs2, Tar)) or (
        isinstance(fs1, Zip) and isinstance(fs2, Zip)
    ):
        return path1 == path2 and fs1._file == fs2._file

    # check all fs with url support
    if isinstance(fs1, fs2.__class__):
        try:
            return fs1.geturl(path1) == fs2.geturl(path2)
        except NoURL:
            pass
    return False


def safe_description(fs: FS, path):
    try:
        if isinstance(fs, SimulationFS):
            return "%s%s" % (str(fs), fspath.abspath(path))
        return fs.getsyspath(path)
    except Exception as e:
        return '{} in "{}"'.format(path, fs)


def ensure_list(inp):
    if not isinstance(inp, list):
        return [inp]
    return inp


def ensure_dict(inp):
    if isinstance(inp, dict):
        return inp
    elif isinstance(inp, str):
        return {inp: {}}
    raise ValueError("Cannot ensure dict: %s" % inp)


def to_args(inp):
    """Convert a argument into a (args, kwargs) tuple.

    >>> to_args(None)
    ([], {})
    >>> to_args('test')
    (['test'], {})
    >>> to_args([1, 2, 3])
    ([1, 2, 3], {})
    >>> to_args({'a': {'b': 'c'}})
    ([], {'a': {'b': 'c'}})
    >>> to_args([[1, 2, [3, 4], [5, 6]]])
    ([1, 2, 3, 4, 5, 6], {})
    """
    if inp is None:
        return ([], {})
    if isinstance(inp, dict):
        return ([], inp)
    return (flatten(ensure_list(inp)), {})


def flatten(arr: List[Any]) -> List[Any]:
    if arr == []:
        return []
    if not isinstance(arr, list):
        return [arr]
    return flatten(arr[0]) + flatten(arr[1:])


def flattened_string_list(x, case_sensitive=True) -> Sequence[str]:
    x = [str(x) for x in flatten(x)]
    if not case_sensitive:
        x = [x.lower() for x in x]
    return x


def flatten_all_lists_in_dict(obj):
    """
    >>> flatten_all_lists_in_dict({1: [[2], [3, {5: [5, 6]}]]})
    {1: [2, 3, {5: [5, 6]}]}
    """
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = flatten_all_lists_in_dict(value)
        return obj
    elif isinstance(obj, list):
        return [flatten_all_lists_in_dict(x) for x in flatten(obj)]
    else:
        return obj


def deep_merge(a: dict, b: dict, *, add_keys=True) -> dict:
    result = deepcopy(a)
    for bk, bv in b.items():
        av = result.get(bk)
        if isinstance(av, dict) and isinstance(bv, dict):
            result[bk] = deep_merge(av, bv, add_keys=add_keys)
        elif (av is not None) or add_keys:
            result[bk] = deepcopy(bv)
    return result


def deep_merge_inplace(base: dict, updates: dict) -> None:
    for bk, bv in updates.items():
        av = base.get(bk)
        if isinstance(av, dict) and isinstance(bv, dict):
            deep_merge_inplace(av, bv)
        else:
            base[bk] = bv



"""
Tests all snippets in the docs and readme like this:

```yaml
rules:
```

To exclude, use shorthand `yml`.
"""

import re

import fs
from conftest import ORGANIZE_DIR
from schema import SchemaError

from organize.actions import ACTIONS
from organize.config import CONFIG_SCHEMA, load_from_string
from organize.filters import FILTERS

DOCS_DIR = str(ORGANIZE_DIR / "docs")

RE_CONFIG = re.compile(r"```yaml\n(?P<config>rules:(?:.*?\n)+?)```", re.MULTILINE)


def test_examples_are_valid():
    docdir = fs.open_fs(DOCS_DIR)
    for f in docdir.walk.files(filter=["*.md"], max_depth=2):
        text = docdir.readtext(f)
        for match in RE_CONFIG.findall(text):
            err = ""
            try:
                config = load_from_string(match)
                CONFIG_SCHEMA.validate(config)
            except SchemaError as e:
                print(f"{f}:\n({match})")
                err = e.autos[-1]
            assert not err


def test_all_filters_documented():
    with fs.open_fs(DOCS_DIR) as docdir:
        filter_docs = docdir.readtext("filters.md")
        for name in FILTERS.keys():
            assert (
                "## {}".format(name) in filter_docs
            ), f"The {name} filter is not documented!"


def test_all_actions_documented():
    with fs.open_fs(DOCS_DIR) as docdir:
        action_docs = docdir.readtext("actions.md")
        for name in ACTIONS.keys():
            assert (
                "## {}".format(name) in action_docs
            ), f"The {name} action is not documented!"



import pytest

from conftest import make_files, read_files
from organize import core


def test_startswith_issue74(testfs):
    # test for issue https://github.com/tfeldmann/organize/issues/74
    make_files(
        testfs,
        {
            "Cálculo_1.pdf": "",
            "Cálculo_2.pdf": "",
            "Calculo.pdf": "",
        },
    )
    config = r"""
        # Cálculo PDF
        rules:
            - locations: "."
              filters:
                - extension:
                    - pdf
                - name:
                    startswith: Cálculo
              actions:
                - move: "Cálculo Integral/Periodo #6/PDF's/"
        """
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "Cálculo Integral": {
            "Periodo #6": {
                "PDF's": {
                    "Cálculo_1.pdf": "",
                    "Cálculo_2.pdf": "",
                }
            }
        },
        "Calculo.pdf": "",
    }


@pytest.mark.skip(reason="TODO")
def test_normalization_regex(testfs):
    make_files(
        testfs,
        {b"Ertra\xcc\x88gnisaufstellung.txt".decode("utf-8"): ""},
    )
    config = (
        b"""
    rules:
      - locations: "."
        filters:
          - regex: 'Ertra\xc3\xa4gnisaufstellung.txt$'
        actions:
          - rename: "found-regex.txt"
    """
    ).decode("utf-8")
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {"found-regex.txt"}


@pytest.mark.skip(reason="TODO")
def test_normalization_filename(testfs):
    make_files(
        testfs,
        {b"Ertr\xcc\x88gnisaufstellung.txt".decode("utf-8"): ""},
    )
    config = (
        b"""
    rules:
      - locations: "."
        filters:
          - filename: "Ertr\xc3\xa4gnisaufstellung"
        actions:
          - rename: "found-regex.txt"
    """
    ).decode("utf-8")
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {"found-regex.txt"}



import os

import fs
from click.testing import CliRunner
from conftest import make_files
from fs.base import FS

from organize import cli

runner = CliRunner()


def test_config_file(tempfs: FS):
    files = ["my-test-file-name.txt", "my-test-file-name.jpg", "other-file.txt"]
    make_files(tempfs, files)
    config = """
    rules:
      - locations: %s
        filters:
          - name: my-test-file-name
        actions:
          - delete
    """
    with fs.open_fs("temp://") as configfs:
        configfs.writetext("config.yaml", config % tempfs.getsyspath("/"))
        args = [configfs.getsyspath("config.yaml")]
        result = runner.invoke(cli.sim, args)
        print(result.output)
        assert set(tempfs.listdir("/")) == set(files)
        result = runner.invoke(cli.run, args)
        print(result.output)
        assert set(tempfs.listdir("/")) == set(["other-file.txt"])


def test_working_dir(tempfs: FS):
    files = ["my-test-file-name.txt", "my-test-file-name.jpg", "other-file.txt"]
    make_files(tempfs, files)
    config = """
    rules:
      - locations: "."
        filters:
          - name: my-test-file-name
        actions:
          - delete
    """
    with fs.open_fs("temp://") as configfs:
        configfs.writetext("config.yaml", config)
        args = [
            configfs.getsyspath("config.yaml"),
            "--working-dir=%s" % tempfs.getsyspath("/"),
        ]
        print(args)
        runner.invoke(cli.sim, args)
        assert set(tempfs.listdir("/")) == set(files)
        runner.invoke(cli.run, args)
        assert set(tempfs.listdir("/")) == set(["other-file.txt"])


def test_with_os_chdir(tempfs: FS):
    files = ["my-test-file-name.txt", "my-test-file-name.jpg", "other-file.txt"]
    make_files(tempfs, files)
    config = """
    rules:
      - locations: "."
        filters:
          - name: my-test-file-name
        actions:
          - delete
    """
    with fs.open_fs("temp://") as configfs:
        configfs.writetext("config.yaml", config)
        args = [
            configfs.getsyspath("config.yaml"),
        ]
        print(args)
        os.chdir(tempfs.getsyspath("/"))
        runner.invoke(cli.sim, args)
        assert set(tempfs.listdir("/")) == set(files)
        runner.invoke(cli.run, args)
        assert set(tempfs.listdir("/")) == set(["other-file.txt"])



import pytest
from fs.base import FS
from conftest import make_files

from organize import core


@pytest.mark.parametrize(
    ("mode", "newline", "result"),
    [
        ("append", "true", "a\nb\nc\n"),
        ("append", "false", "abc"),
        ("prepend", "true", "c\nb\na\n"),
        ("prepend", "false", "cba"),
        ("overwrite", "true", "c\n"),
        ("overwrite", "false", "c"),
    ],
)
def test_write(testfs: FS, mode, newline, result):
    files = ["a.txt", "b.txt", "c.txt"]
    make_files(testfs, files)
    config = """
    rules:
      - locations: "."
        filters:
          - name: "a"
        actions:
          - write:
              text: "{text}"
              path: "out.txt"
              mode: {mode}
              newline: {newline}
      - locations: "."
        filters:
          - name: "b"
        actions:
          - write:
              text: "{text}"
              path: "out.txt"
              mode: {mode}
              newline: {newline}
      - locations: "."
        filters:
          - name: "c"
        actions:
          - write:
              text: "{text}"
              path: "out.txt"
              mode: {mode}
              newline: {newline}
    """.format(
        text="{name}", mode=mode, newline=newline
    )
    core.run(config, simulate=False, working_dir=testfs)
    assert testfs.readtext("out.txt") == result



from textwrap import indent
from typing import Any, Dict, NamedTuple, Union

from schema import Optional, Or, Schema

from organize.console import pipeline_error, pipeline_message


class FilterResult(NamedTuple):
    matches: bool
    updates: dict


class Filter:
    name = None  # type: Union[str, None]
    arg_schema = None  # type: Union[Schema, None]
    schema_support_instance_without_args = False

    @classmethod
    def get_name(cls):
        if cls.name:
            return cls.name
        return cls.__name__.lower()

    @classmethod
    def get_name_schema(cls):
        return Schema(
            Or("not " + cls.get_name(), cls.get_name()),
            name=cls.get_name(),
            description=cls.get_description(),
        )

    @classmethod
    def get_schema(cls):
        name = cls.get_name_schema()

        if cls.arg_schema:
            arg_schema = cls.arg_schema
        else:
            arg_schema = Or(
                str,
                [str],
                Schema({}, ignore_extra_keys=True),
            )

        if cls.schema_support_instance_without_args:
            return Or(name, {name: arg_schema})
        return {
            name: arg_schema,
        }

    @classmethod
    def get_description(cls):
        """the first line of the class docstring"""
        return cls.__doc__.splitlines()[0]

    def run(self, **kwargs: Dict) -> FilterResult:
        return self.pipeline(dict(kwargs))

    def pipeline(self, args: dict) -> FilterResult:
        raise NotImplementedError

    def print(self, *msg: str) -> None:
        """print a message for the user"""
        text = " ".join(str(x) for x in msg)
        for line in text.splitlines():
            pipeline_message(self.get_name(), line)

    def print_error(self, msg: str):
        for line in msg.splitlines():
            pipeline_error(self.get_name(), line)

    def set_logic(self, inverted=False):
        self.inverted = inverted

    def __str__(self) -> str:
        """Return filter name and properties"""
        return self.get_name()

    def __repr__(self) -> str:
        return "<%s>" % str(self)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__



from typing import Dict, Type

from .created import Created
from .date_added import DateAdded
from .date_lastused import DateLastUsed
from .duplicate import Duplicate
from .empty import Empty
from .exif import Exif
from .extension import Extension
from .filecontent import FileContent
from .filter import Filter
from .hash import Hash
from .lastmodified import LastModified
from .macos_tags import MacOSTags
from .mimetype import MimeType
from .name import Name
from .python import Python
from .regex import Regex
from .size import Size

FILTERS: Dict[str, Type[Filter]]
FILTERS = {
    Created.name: Created,
    DateAdded.name: DateAdded,
    DateLastUsed.name: DateLastUsed,
    Duplicate.name: Duplicate,
    Empty.name: Empty,
    Exif.name: Exif,
    Extension.name: Extension,
    FileContent.name: FileContent,
    Hash.name: Hash,
    Name.name: Name,
    Size.name: Size,
    LastModified.name: LastModified,
    MacOSTags.name: MacOSTags,
    MimeType.name: MimeType,
    Python.name: Python,
    Regex.name: Regex,
}



from typing import Callable, Union

from fs import errors, open_fs
from fs.base import FS
from fs.copy import copy_dir, copy_file
from fs.opener.errors import OpenerError
from fs.path import dirname
from schema import Optional, Or

from organize.utils import SimulationFS, Template, safe_description

from .action import Action
from ._conflict_resolution import CONFLICT_OPTIONS, check_conflict, dst_from_options


class Copy(Action):

    """Copy a file or dir to a new location.

    If the specified path does not exist it will be created.

    Args:
        dest (str):
            The destination where the file / dir should be copied to.
            If `dest` ends with a slash, it is assumed to be a target directory
            and the file / dir will be copied into `dest` and keep its name.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

        filesystem (str):
            (Optional) A pyfilesystem opener url of the filesystem you want to copy to.
            If this is not given, the local filesystem is used.

    The next action will work with the created copy.
    """

    name = "copy"
    arg_schema = Or(
        str,
        {
            "dest": str,
            Optional("on_conflict"): Or(*CONFLICT_OPTIONS),
            Optional("rename_template"): str,
            Optional("filesystem"): object,
        },
    )

    def __init__(
        self,
        dest: str,
        on_conflict="rename_new",
        rename_template="{name} {counter}{extension}",
        filesystem: Union[FS, str, None] = None,
    ) -> None:
        if on_conflict not in CONFLICT_OPTIONS:
            raise ValueError(
                "on_conflict must be one of %s" % ", ".join(CONFLICT_OPTIONS)
            )

        self.dest = Template.from_string(dest)
        self.conflict_mode = on_conflict
        self.rename_template = Template.from_string(rename_template)
        self.filesystem = filesystem or self.Meta.default_filesystem

    def pipeline(self, args: dict, simulate: bool):
        src_fs = args["fs"]  # type: FS
        src_path = args["fs_path"]

        # should we copy a dir or a file?
        copy_action: Callable[[FS, str, FS, str], None]
        if src_fs.isdir(src_path):
            copy_action = copy_dir
        elif src_fs.isfile(src_path):
            copy_action = copy_file

        dst_fs, dst_path = dst_from_options(
            src_path=src_path,
            dest=self.dest,
            filesystem=self.filesystem,
            args=args,
        )

        # check for conflicts
        skip, dst_path = check_conflict(
            src_fs=src_fs,
            src_path=src_path,
            dst_fs=dst_fs,
            dst_path=dst_path,
            conflict_mode=self.conflict_mode,
            rename_template=self.rename_template,
            simulate=simulate,
            print=self.print,
        )

        try:
            dst_fs = open_fs(dst_fs, create=False, writeable=True)
        except (errors.CreateFailed, OpenerError):
            if not simulate:
                dst_fs = open_fs(dst_fs, create=True, writeable=True)
            else:
                dst_fs = SimulationFS(dst_fs)

        if not skip:
            self.print("Copy to %s" % safe_description(dst_fs, dst_path))
            if not simulate:
                dst_fs.makedirs(dirname(dst_path), recreate=True)
                copy_action(src_fs, src_path, dst_fs, dst_path)

        # the next action should work with the newly created copy
        return {
            "fs": dst_fs,
            "fs_path": dst_path,
        }

    def __str__(self) -> str:
        return "Copy(dest=%s, conflict_mode=%s)" % (self.dest, self.conflict_mode)



import fs

from organize.filters import Name


def test_name_startswith():
    name = Name(startswith="begin")
    assert name.matches("beginhere")
    assert not name.matches(".beginhere")
    assert not name.matches("herebegin")


def test_name_contains():
    name = Name(contains="begin")
    assert name.matches("beginhere")
    assert name.matches(".beginhere")
    assert name.matches("herebegin")
    assert not name.matches("other")


def test_name_endswith():
    name = Name(endswith="end")
    assert name.matches("hereend")
    assert name.matches("end")
    assert not name.matches("theendishere")


def test_name_multiple():
    name = Name(startswith="begin", contains="con", endswith="end")
    assert name.matches("begin_somethgin_con_end")
    assert not name.matches("beginend")
    assert not name.matches("begincon")
    assert not name.matches("conend")
    assert name.matches("beginconend")


def test_name_case():
    name = Name(startswith="star", contains="con", endswith="end", case_sensitive=False)
    assert name.matches("STAR_conEnD")
    assert not name.matches("STAREND")
    assert not name.matches("STARCON")
    assert not name.matches("CONEND")
    assert name.matches("STARCONEND")


def test_name_list():
    name = Name(
        startswith="_",
        contains=["1", "A", "3", "6"],
        endswith=["5", "6"],
        case_sensitive=False,
    )
    assert name.matches("_15")
    assert name.matches("_A5")
    assert name.matches("_A6")
    assert name.matches("_a6")
    assert name.matches("_35")
    assert name.matches("_36")
    assert name.matches("_somethinga56")
    assert name.matches("_6")
    assert not name.matches("")
    assert not name.matches("a_5")


def test_name_list_case_sensitive():
    name = Name(
        startswith="_",
        contains=["1", "A", "3", "7"],
        endswith=["5", "6"],
        case_sensitive=True,
    )
    assert name.matches("_15")
    assert name.matches("_A5")
    assert name.matches("_A6")
    assert not name.matches("_a6")
    assert name.matches("_35")
    assert name.matches("_36")
    assert name.matches("_somethingA56")
    assert not name.matches("_6")
    assert not name.matches("_a5")
    assert not name.matches("-A5")
    assert not name.matches("")
    assert not name.matches("_a5")


def test_name_match():
    with fs.open_fs("mem://") as mem:
        p = "Invoice_RE1001_2021_01_31"
        fs_path = p + ".txt"
        mem.touch(fs_path)
        fn = Name("Invoice_*_{year:int}_{month}_{day}")
        assert fn.matches(p)
        result = fn.run(fs=mem, fs_path=fs_path)
        assert result.matches
        assert result.updates == {"name": {"year": 2021, "month": "01", "day": "31"}}


def test_name_match_case_insensitive():
    with fs.open_fs("mem://") as mem:
        p = "UPPER_MiXed_lower"
        fs_path = p + ".txt"
        mem.touch(fs_path)
        case = Name("upper_{m1}_{m2}", case_sensitive=True)
        icase = Name("upper_{m1}_{m2}", case_sensitive=False)
        assert icase.matches(p)
        result = icase.run(fs=mem, fs_path=fs_path)
        assert result.matches
        assert result.updates == {"name": {"m1": "MiXed", "m2": "lower"}}
        assert not case.matches(p)



import pytest
from fs import open_fs
from schema import SchemaError

from organize import config, core


def validate_config(string: str):
    conf = config.load_from_string(string)
    conf = config.cleanup(conf)
    config.validate(conf)
    return conf


def test_basic():
    STR = """
    rules:
    - locations: '~/'
      filters:
      - extension:
        - jpg
        - png
      - extension: txt
      actions:
      - move:
          dest: '~/New Folder'
      - echo: 'Moved {path}/{extension.upper()}'
    - locations:
      - path: '~/test1'
        ignore_errors: true
      actions:
      - shell:
          cmd: 'say {path.stem}'
    """
    validate_config(STR)


def test_yaml_ref():
    STR = """
    media: &media
      - wav
      - png

    all_folders: &all
      - "~"
      - "/"

    rules:
      - locations: *all
        filters:
          - extension: *media
          - extension:
            - *media
            - jpg
          - lastmodified:
              days: 10
        actions:
          - echo: 'Hello World'
      - locations:
          - *all
          - path: /more/more
            ignore_errors: true
        actions:
          - trash
    """
    validate_config(STR)


def test_error_filter_dict():
    STR = """
    rules:
    - locations: '/'
      filters:
        extension: 'jpg'
      actions:
      - trash
    """
    with pytest.raises(SchemaError):
        validate_config(STR)


# def test_error_action_dict():
#     conf = Config.from_string(
#         """
#     rules:
#     - folders: '/'
#       filters:
#       - extension: 'jpg'
#       actions:
#         Trash
#     """
#     )
#     with pytest.raises(Config.ActionsNoListError):
#         _ = conf.rules


# def test_empty_filters():
#     conf = """
#     rules:
#       - folders: '/'
#         filters:
#         actions:
#           - trash
#       - folders: '~/'
#         actions:
#           - trash
#     """
#     assert Config.from_string(conf).rules == [
#         Rule(
#             folders=["/"],
#             filters=[],
#             actions=[Trash()],
#             subfolders=False,
#             system_files=False,
#         ),
#         Rule(
#             folders=["~/"],
#             filters=[],
#             actions=[Trash()],
#             subfolders=False,
#             system_files=False,
#         ),
#     ]


# @pytest.mark.skip
# def test_flatten_filters_and_actions():
#     config = """
#     folder_aliases:
#       Downloads: &downloads ~/Downloads/
#       Payables_due: &payables_due ~/PayablesDue/
#       Payables_paid: &payables_paid ~/Accounting/Expenses/
#       Receivables_due: &receivables_due ~/Receivables/
#       Receivables_paid: &receivables_paid ~/Accounting/Income/

#     defaults:
#       filters: &default_filters
#         - extension: pdf
#         - filecontent: '(?P<date>...)'
#       actions: &default_actions
#         - echo: 'Dated: {filecontent.date}'
#         - echo: 'Stem of filename: {filecontent.stem}'
#       post_actions: &default_sorting
#         - rename: '{python.timestamp}-{filecontent.stem}.{extension.lower}'
#         - move: '{path.parent}/{python.quarter}/'

#     rules:
#       - folders: *downloads
#         filters:
#           - *default_filters
#           - filecontent: 'Due Date' # regex to id as payable
#           - filecontent: '(?P<stem>...)' # regex to extract supplier
#         actions:
#           - *default_actions
#           - move: *payables_due
#           - *default_sorting

#       - folders: *downloads
#         filters:
#           - *default_filters
#           - filecontent: 'Account: 000000000' # regex to id as receivables due
#           - filecontent: '(?P<stem>...)' # regex to extract customer
#         actions:
#           - *default_actions
#           - move: *receivables_due
#           - *default_sorting

#       - folders: *downloads
#         filters:
#           - *default_filters
#           - filecontent: 'PAID' # regex to id as receivables paid
#           - filecontent: '(?P<stem>...)' # regex to extract customer
#           - filecontent: '(?P<paid>...)' # regex to extract date paid
#           - filename:
#               startswith: 2020
#         actions:
#           - *default_actions
#           - move: *receivables_paid
#           - *default_sorting
#           - rename: '{filecontent.paid}_{filecontent.stem}.{extension}'
#     """
#     conf = Config.from_string(config)
#     assert conf.rules == [
#         Rule(
#             folders=["~/Downloads/"],
#             filters=[
#                 # default_filters
#                 Extension("pdf"),
#                 FileContent(expr="(?P<date>...)"),
#                 # added filters
#                 FileContent(expr="Due Date"),
#                 FileContent(expr="(?P<stem>...)"),
#             ],
#             actions=[
#                 # default_actions
#                 Echo(msg="Dated: {filecontent.date}"),
#                 Echo(msg="Stem of filename: {filecontent.stem}"),
#                 # added actions
#                 Move(dest="~/PayablesDue/", overwrite=False),
#                 # default_sorting
#                 Rename(
#                     name="{python.timestamp}-{filecontent.stem}.{extension.lower}",
#                     overwrite=False,
#                 ),
#                 Move(dest="{path.parent}/{python.quarter}/", overwrite=False),
#             ],
#             subfolders=False,
#             system_files=False,
#         ),
#         Rule(
#             folders=["~/Downloads/"],
#             filters=[
#                 # default_filters
#                 Extension("pdf"),
#                 FileContent(expr="(?P<date>...)"),
#                 # added filters
#                 FileContent(expr="Account: 000000000"),
#                 FileContent(expr="(?P<stem>...)"),
#             ],
#             actions=[
#                 # default_actions
#                 Echo(msg="Dated: {filecontent.date}"),
#                 Echo(msg="Stem of filename: {filecontent.stem}"),
#                 # added actions
#                 Move(dest="~/Receivables/", overwrite=False),
#                 # default_sorting
#                 Rename(
#                     name="{python.timestamp}-{filecontent.stem}.{extension.lower}",
#                     overwrite=False,
#                 ),
#                 Move(dest="{path.parent}/{python.quarter}/", overwrite=False),
#             ],
#             subfolders=False,
#             system_files=False,
#         ),
#         Rule(
#             folders=["~/Downloads/"],
#             filters=[
#                 # default_filters
#                 Extension("pdf"),
#                 FileContent(expr="(?P<date>...)"),
#                 # added filters
#                 FileContent(expr="PAID"),
#                 FileContent(expr="(?P<stem>...)"),
#                 FileContent(expr="(?P<paid>...)"),
#                 Filename(startswith="2020"),
#             ],
#             actions=[
#                 # default_actions
#                 Echo(msg="Dated: {filecontent.date}"),
#                 Echo(msg="Stem of filename: {filecontent.stem}"),
#                 # added actions
#                 Move(dest="~/Accounting/Income/", overwrite=False),
#                 # default_sorting
#                 Rename(
#                     name="{python.timestamp}-{filecontent.stem}.{extension.lower}",
#                     overwrite=False,
#                 ),
#                 Move(dest="{path.parent}/{python.quarter}/", overwrite=False),
#                 # added actions
#                 Rename(
#                     name="{filecontent.paid}_{filecontent.stem}.{extension}",
#                     overwrite=False,
#                 ),
#             ],
#             subfolders=False,
#             system_files=False,
#         ),
#     ]



from fs.base import FS

from .filter import Filter, FilterResult


class Empty(Filter):

    """Finds empty dirs and files"""

    name = "empty"

    @classmethod
    def get_schema(cls):
        return cls.get_name_schema()

    def pipeline(self, args: dict) -> FilterResult:
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]  # type: str

        if fs.isdir(fs_path):
            result = fs.isempty(fs_path)
        else:
            result = fs.getsize(fs_path) == 0

        return FilterResult(matches=result, updates={})

    def __str__(self) -> str:
        return "Empty()"



from datetime import datetime, timedelta
from typing import Union

from schema import Optional, Or

from .filter import Filter, FilterResult
from .utils import age_condition_applies


class TimeFilter(Filter):

    schema_support_instance_without_args = True
    arg_schema = {
        Optional("mode"): Or("older", "newer"),
        Optional("years"): int,
        Optional("months"): int,
        Optional("weeks"): int,
        Optional("days"): int,
        Optional("hours"): int,
        Optional("minutes"): int,
        Optional("seconds"): int,
    }

    def __init__(
        self,
        years=0,
        months=0,
        weeks=0,
        days=0,
        hours=0,
        minutes=0,
        seconds=0,
        mode="older",
    ):
        self.age = timedelta(
            weeks=52 * years + 4 * months + weeks,  # quick and a bit dirty
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
        )
        self.mode = mode.strip().lower()
        if self.mode not in ("older", "newer"):
            raise ValueError("Unknown option for 'mode': must be 'older' or 'newer'.")

    def matches_datetime(self, dt: datetime) -> bool:
        match = True
        if self.age.total_seconds():
            if not dt:
                match = False
            else:
                match = age_condition_applies(
                    dt=dt,
                    age=self.age,
                    mode=self.mode,
                    reference=datetime.now(),
                )
        return match

    def pipeline(self, args: dict) -> FilterResult:
        dt = self.get_datetime(args)
        if dt is None:
            return FilterResult(matches=False, updates={})
        dt = dt.astimezone()

        match = self.matches_datetime(dt)
        return FilterResult(
            matches=match,
            updates={self.get_name(): dt},
        )

    def get_datetime(self, args: dict) -> Union[datetime, None]:
        raise NotImplemented



from typing import Dict, Union

from fs import errors
from fs.base import FS
from fs.opener import open_fs
from fs.opener.errors import OpenerError
from fs.path import basename, dirname

from organize.utils import SimulationFS, expand_args


def open_create_fs_path(fs: Union[str, FS], path: str, args: Dict, simulate: bool):
    if fs:
        if isinstance(fs, str):
            dst_fs = expand_args(fs, args)
        else:
            dst_fs = fs
    else:
        dst_fs = dirname(path)
    dst_path = basename(path)
    try:
        dst_fs = open_fs(dst_fs, create=False, writeable=True)
    except (errors.CreateFailed, OpenerError):
        if not simulate:
            dst_fs = open_fs(dst_fs, create=True, writeable=True)
        else:
            dst_fs = SimulationFS(dst_fs)
    return dst_fs, dst_path



import collections
from pathlib import Path
from typing import Any, DefaultDict, Dict, Mapping, Optional, Union

import exifread

from .filter import Filter, FilterResult

ExifDict = Mapping[str, Union[str, Mapping[str, str]]]


class Exif(Filter):
    """Filter by image EXIF data

    The `exif` filter can be used as a filter as well as a way to get exif information
    into your actions.

    :returns:
        ``{exif}`` -- a dict of all the collected exif inforamtion available in the
        file. Typically it consists of the following tags (if present in the file):

        - ``{exif.image}`` -- information related to the main image
        - ``{exif.exif}`` -- Exif information
        - ``{exif.gps}`` -- GPS information
        - ``{exif.interoperability}`` -- Interoperability information
    """

    name = "exif"
    arg_schema = None
    schema_support_instance_without_args = True

    def __init__(self, *required_tags: str, **tag_filters: str) -> None:
        self.args = required_tags  # expected exif keys
        self.kwargs = tag_filters  # exif keys with expected values

    def category_dict(self, tags: Mapping[str, str]) -> ExifDict:
        result = collections.defaultdict(dict)  # type: DefaultDict
        for key, value in tags.items():
            if " " in key:
                category, field = key.split(" ", maxsplit=1)
                result[category][field] = value
            else:
                result[key] = value
        return dict(result)

    def matches(self, exiftags: dict) -> bool:
        if not exiftags:
            return False
        tags = {k.lower(): v for k, v in exiftags.items()}

        # no match if expected tag is not found
        normkey = lambda k: k.replace(".", " ").lower()
        for key in self.args:
            if normkey(key) not in tags:
                return False
        # no match if tag has not expected value
        for key, value in self.kwargs.items():
            key = normkey(key)
            if not (key in tags and tags[key].lower() == value.lower()):
                return False
        return True

    def pipeline(self, args: dict) -> FilterResult:
        fs = args["fs"]
        fs_path = args["fs_path"]
        with fs.openbin(fs_path) as f:
            exiftags = exifread.process_file(f, details=False)

        tags = {k.lower(): v.printable for k, v in exiftags.items()}
        matches = self.matches(tags)
        exif_result = self.category_dict(tags)

        return FilterResult(
            matches=matches,
            updates={self.get_name(): exif_result},
        )

    def __str__(self) -> str:
        return "EXIF(%s)" % ", ".join("%s=%s" % (k, v) for k, v in self.kwargs.items())



import fs
from conftest import make_files, read_files

from organize.core import run


def test_rename_issue51(testfs):
    # test for issue https://github.com/tfeldmann/organize/issues/51
    files = {
        "19asd_WF_test2.PDF": "",
        "other.pdf": "",
        "18asd_WFX_test2.pdf": "",
    }
    make_files(testfs, files)

    CONFIG = """
    rules:
      - locations: "./"
        filters:
          - extension
          - name:
              startswith: "19"
              contains:
                - "_WF_"
        actions:
          - rename: "{name}_unread.{extension.lower()}"
          - copy:
              dest: "copy/"
    """
    run(CONFIG, simulate=False, working_dir=testfs)
    result = read_files(testfs)
    assert result == {
        "copy": {
            "19asd_WF_test2_unread.pdf": "",
        },
        "19asd_WF_test2_unread.pdf": "",
        "other.pdf": "",
        "18asd_WFX_test2.pdf": "",
    }


def test_rename_folders(testfs):
    config = """
    rules:
      - name: "Renaming DVD folders"
        locations: "/"
        targets: dirs
        filters:
          - name:
              contains: DVD
        actions:
          - rename:
              name: "{name.replace('[DVD] ','').replace(' [1080p]','').replace(' ', '_')}"
    """
    files = {
        "[DVD] Best Of Video 1080 [1080p]": {
            "[DVD] Best Of Video 1080 [1080p]": "",
            "Metadata": "",
        },
        "[DVD] This Is A Title [1080p]": {
            "[DVD] This Is A Title [1080p]": "",
            "Metadata": "",
        },
    }
    make_files(testfs, files)
    run(rules=config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "Best_Of_Video_1080": {
            "[DVD] Best Of Video 1080 [1080p]": "",
            "Metadata": "",
        },
        "This_Is_A_Title": {
            "[DVD] This Is A Title [1080p]": "",
            "Metadata": "",
        },
    }


def test_rename_in_subfolders(testfs):
    config = """
    rules:
      - locations: "/"
        subfolders: true
        filters:
          - name:
              contains: RENAME
          - extension
        actions:
          - rename: "DONE.{extension}"
    """
    files = {
        "folder": {
            "FIRST-RENAME.pdf": "",
            "Other": "",
        },
        "Another folder": {
            "Subfolder": {
                "RENAME-ME_TOO.txt": "",
            },
            "Metadata": "",
        },
    }
    make_files(testfs, files)
    run(rules=config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "folder": {
            "DONE.pdf": "",
            "Other": "",
        },
        "Another folder": {
            "Subfolder": {
                "DONE.txt": "",
            },
            "Metadata": "",
        },
    }


def test_filename_move(tempfs):
    config = """
        rules:
        - locations: "."
          filters:
            - extension
          actions:
            - rename: '{path.stem}{path.stem}.{extension.lower()}'
    """
    make_files(tempfs, {"test.PY": ""})
    run(rules=config, simulate=False, working_dir=tempfs)
    assert read_files(tempfs) == {"testtest.py": ""}
